u"""
oauthlib.oauth1.rfc5849
~~~~~~~~~~~~~~

This module is an implementation of various logic needed
for signing and checking OAuth 1.0 RFC 5849 requests.
"""
from __future__ import absolute_import, unicode_literals
import base64
import hashlib
import sys
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from . import parameters, signature
from . import common
SIGNATURE_HMAC_SHA1 = u'HMAC-SHA1'
SIGNATURE_HMAC_SHA256 = u'HMAC-SHA256'
SIGNATURE_HMAC = SIGNATURE_HMAC_SHA1
SIGNATURE_RSA = u'RSA-SHA1'
SIGNATURE_PLAINTEXT = u'PLAINTEXT'
SIGNATURE_METHODS = (SIGNATURE_HMAC_SHA1,
 SIGNATURE_HMAC_SHA256,
 SIGNATURE_RSA,
 SIGNATURE_PLAINTEXT)
SIGNATURE_TYPE_AUTH_HEADER = u'AUTH_HEADER'
SIGNATURE_TYPE_QUERY = u'QUERY'
SIGNATURE_TYPE_BODY = u'BODY'
CONTENT_TYPE_FORM_URLENCODED = u'application/x-www-form-urlencoded'

class Client(object):
    u"""A client used to sign OAuth 1.0 RFC 5849 requests."""
    SIGNATURE_METHODS = {SIGNATURE_HMAC_SHA1: signature.sign_hmac_sha1_with_client,
     SIGNATURE_HMAC_SHA256: signature.sign_hmac_sha256_with_client,
     SIGNATURE_RSA: signature.sign_rsa_sha1_with_client,
     SIGNATURE_PLAINTEXT: signature.sign_plaintext_with_client}

    @classmethod
    def register_signature_method(cls, method_name, method_callback):
        cls.SIGNATURE_METHODS[method_name] = method_callback

    def __init__(self, client_key, client_secret = None, resource_owner_key = None, resource_owner_secret = None, callback_uri = None, signature_method = SIGNATURE_HMAC_SHA1, signature_type = SIGNATURE_TYPE_AUTH_HEADER, rsa_key = None, verifier = None, realm = None, encoding = u'utf-8', decoding = None, nonce = None, timestamp = None):
        u"""Create an OAuth 1 client.
        
        :param client_key: Client key (consumer key), mandatory.
        :param resource_owner_key: Resource owner key (oauth token).
        :param resource_owner_secret: Resource owner secret (oauth token secret).
        :param callback_uri: Callback used when obtaining request token.
        :param signature_method: SIGNATURE_HMAC, SIGNATURE_RSA or SIGNATURE_PLAINTEXT.
        :param signature_type: SIGNATURE_TYPE_AUTH_HEADER (default),
                               SIGNATURE_TYPE_QUERY or SIGNATURE_TYPE_BODY
                               depending on where you want to embed the oauth
                               credentials.
        :param rsa_key: RSA key used with SIGNATURE_RSA.
        :param verifier: Verifier used when obtaining an access token.
        :param realm: Realm (scope) to which access is being requested.
        :param encoding: If you provide non-unicode input you may use this
                         to have oauthlib automatically convert.
        :param decoding: If you wish that the returned uri, headers and body
                         from sign be encoded back from unicode, then set
                         decoding to your preferred encoding, i.e. utf-8.
        :param nonce: Use this nonce instead of generating one. (Mainly for testing)
        :param timestamp: Use this timestamp instead of using current. (Mainly for testing)
        """
        encode = lambda x: (common.to_unicode(x, encoding) if encoding else x)
        self.client_key = encode(client_key)
        self.client_secret = encode(client_secret)
        self.resource_owner_key = encode(resource_owner_key)
        self.resource_owner_secret = encode(resource_owner_secret)
        self.signature_method = encode(signature_method)
        self.signature_type = encode(signature_type)
        self.callback_uri = encode(callback_uri)
        self.rsa_key = encode(rsa_key)
        self.verifier = encode(verifier)
        self.realm = encode(realm)
        self.encoding = encode(encoding)
        self.decoding = encode(decoding)
        self.nonce = encode(nonce)
        self.timestamp = encode(timestamp)

    def __repr__(self):
        attrs = vars(self).copy()
        attrs[u'client_secret'] = u'****' if attrs[u'client_secret'] else None
        attrs[u'rsa_key'] = u'****' if attrs[u'rsa_key'] else None
        attrs[u'resource_owner_secret'] = u'****' if attrs[u'resource_owner_secret'] else None
        attribute_str = u', '.join((u'%s=%s' % (k, v) for k, v in list(attrs.items())))
        return u'<%s %s>' % (self.__class__.__name__, attribute_str)

    def get_oauth_signature(self, request):
        u"""Get an OAuth signature to be used in signing a request
        
        To satisfy `section 3.4.1.2`_ item 2, if the request argument's
        headers dict attribute contains a Host item, its value will
        replace any netloc part of the request argument's uri attribute
        value.
        
        .. _`section 3.4.1.2`: https://tools.ietf.org/html/rfc5849#section-3.4.1.2
        """
        if self.signature_method == SIGNATURE_PLAINTEXT:
            return signature.sign_plaintext(self.client_secret, self.resource_owner_secret)
        else:
            uri, headers, body = self._render(request)
            collected_params = signature.collect_parameters(uri_query=urlparse.urlparse(uri).query, body=body, headers=headers)
            normalized_params = signature.normalize_parameters(collected_params)
            normalized_uri = signature.normalize_base_string_uri(uri, headers.get(u'Host', None))
            base_string = signature.construct_base_string(request.http_method, normalized_uri, normalized_params)
            if self.signature_method not in self.SIGNATURE_METHODS:
                raise ValueError(u'Invalid signature method.')
            sig = self.SIGNATURE_METHODS[self.signature_method](base_string, self)
            return sig

    def get_oauth_params(self, request):
        u"""Get the basic OAuth parameters to be used in generating a signature.
        """
        nonce = common.generate_nonce() if self.nonce is None else self.nonce
        timestamp = common.generate_timestamp() if self.timestamp is None else self.timestamp
        params = [(u'oauth_nonce', nonce),
         (u'oauth_timestamp', timestamp),
         (u'oauth_version', u'1.0'),
         (u'oauth_signature_method', self.signature_method),
         (u'oauth_consumer_key', self.client_key)]
        if self.resource_owner_key:
            params.append((u'oauth_token', self.resource_owner_key))
        if self.callback_uri:
            params.append((u'oauth_callback', self.callback_uri))
        if self.verifier:
            params.append((u'oauth_verifier', self.verifier))
        content_type = request.headers.get(u'Content-Type', None)
        content_type_eligible = content_type and content_type.find(u'application/x-www-form-urlencoded') < 0
        if request.body is not None and content_type_eligible:
            params.append((u'oauth_body_hash', base64.b64encode(hashlib.sha1(request.body.encode(u'utf-8')).digest()).decode(u'utf-8')))
        return params

    def _render(self, request, formencode = False, realm = None):
        u"""Render a signed request according to signature type
        
        Returns a 3-tuple containing the request URI, headers, and body.
        
        If the formencode argument is True and the body contains parameters, it
        is escaped and returned as a valid formencoded string.
        """
        uri, headers, body = request.uri, request.headers, request.body
        if self.signature_type == SIGNATURE_TYPE_AUTH_HEADER:
            headers = parameters.prepare_headers(request.oauth_params, request.headers, realm=realm)
        elif self.signature_type == SIGNATURE_TYPE_BODY and request.decoded_body is not None:
            body = parameters.prepare_form_encoded_body(request.oauth_params, request.decoded_body)
            if formencode:
                body = common.urlencode(body)
            headers[u'Content-Type'] = u'application/x-www-form-urlencoded'
        elif self.signature_type == SIGNATURE_TYPE_QUERY:
            uri = parameters.prepare_request_uri_query(request.oauth_params, request.uri)
        else:
            raise ValueError(u'Unknown signature type specified.')
        return (uri, headers, body)

    def sign(self, uri, http_method = u'GET', body = None, headers = None, realm = None):
        u"""Sign a request
        
        Signs an HTTP request with the specified parts.
        
        Returns a 3-tuple of the signed request's URI, headers, and body.
        Note that http_method is not returned as it is unaffected by the OAuth
        signing process. Also worth noting is that duplicate parameters
        will be included in the signature, regardless of where they are
        specified (query, body).
        
        The body argument may be a dict, a list of 2-tuples, or a formencoded
        string. The Content-Type header must be 'application/x-www-form-urlencoded'
        if it is present.
        
        If the body argument is not one of the above, it will be returned
        verbatim as it is unaffected by the OAuth signing process. Attempting to
        sign a request with non-formencoded data using the OAuth body signature
        type is invalid and will raise an exception.
        
        If the body does contain parameters, it will be returned as a properly-
        formatted formencoded string.
        
        Body may not be included if the http_method is either GET or HEAD as
        this changes the semantic meaning of the request.
        
        All string data MUST be unicode or be encoded with the same encoding
        scheme supplied to the Client constructor, default utf-8. This includes
        strings inside body dicts, for example.
        """
        request = common.Request(uri, http_method, body, headers, encoding=self.encoding)
        content_type = request.headers.get(u'Content-Type', None)
        multipart = content_type and content_type.startswith(u'multipart/')
        should_have_params = content_type == CONTENT_TYPE_FORM_URLENCODED
        has_params = request.decoded_body is not None
        if multipart and has_params:
            raise ValueError(u'Headers indicate a multipart body but body contains parameters.')
        elif should_have_params and not has_params:
            raise ValueError(u'Headers indicate a formencoded body but body was not decodable.')
        elif not should_have_params and has_params:
            raise ValueError(u'Body contains parameters but Content-Type header was {0} instead of {1}'.format(content_type or u'not set', CONTENT_TYPE_FORM_URLENCODED))
        elif self.signature_type == SIGNATURE_TYPE_BODY and not (should_have_params and has_params and not multipart):
            raise ValueError(u'Body signatures may only be used with form-urlencoded content')
        elif http_method.upper() in (u'GET', u'HEAD') and has_params:
            raise ValueError(u'GET/HEAD requests should not include body.')
        request.oauth_params = self.get_oauth_params(request)
        request.oauth_params.append((u'oauth_signature', self.get_oauth_signature(request)))
        uri, headers, body = self._render(request, formencode=True, realm=realm or self.realm)
        if self.decoding:
            uri = uri.encode(self.decoding)
            body = body.encode(self.decoding) if body else body
            new_headers = {}
            for k, v in list(headers.items()):
                new_headers[k.encode(self.decoding)] = v.encode(self.decoding)

            headers = new_headers
        return (uri, headers, body)