u"""
oauthlib.common
~~~~~~~~~~~~~~

This module provides data structures and utilities common
to all implementations of OAuth.
"""
from __future__ import absolute_import, unicode_literals
import collections
import datetime
import re
import sys
import time
try:
    from secrets import randbits
    from secrets import SystemRandom
except ImportError:
    from random import getrandbits as randbits
    from random import SystemRandom

try:
    from urllib import quote as _quote
    from urllib import unquote as _unquote
    from urllib import urlencode as _urlencode
except ImportError:
    from urllib.parse import quote as _quote
    from urllib.parse import unquote as _unquote
    from urllib.parse import urlencode as _urlencode

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

UNICODE_ASCII_CHARACTER_SET = u'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
CLIENT_ID_CHARACTER_SET = u' !"#$%&\\\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}'
SANITIZE_PATTERN = re.compile(u'([^&;]*(?:password|token)[^=]*=)[^&;]+', re.IGNORECASE)
INVALID_HEX_PATTERN = re.compile(u'%[^0-9A-Fa-f]|%[0-9A-Fa-f][^0-9A-Fa-f]')
always_safe = u'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_.-'
PY3 = sys.version_info[0] == 3
if PY3:
    unicode_type = str
else:
    unicode_type = unicode

def quote(s, safe = '/'):
    s = s.encode(u'utf-8') if isinstance(s, unicode_type) else s
    s = _quote(s, safe)
    if isinstance(s, bytes):
        s = s.decode(u'utf-8')
    return s


def unquote(s):
    s = _unquote(s)
    if isinstance(s, bytes):
        s = s.decode(u'utf-8')
    return s


def urlencode(params):
    utf8_params = encode_params_utf8(params)
    urlencoded = _urlencode(utf8_params)
    if isinstance(urlencoded, unicode_type):
        return urlencoded
    else:
        return urlencoded.decode(u'utf-8')


def encode_params_utf8(params):
    u"""Ensures that all parameters in a list of 2-element tuples are encoded to
    bytestrings using UTF-8
    """
    encoded = []
    for k, v in params:
        encoded.append((k.encode(u'utf-8') if isinstance(k, unicode_type) else k, v.encode(u'utf-8') if isinstance(v, unicode_type) else v))

    return encoded


def decode_params_utf8(params):
    u"""Ensures that all parameters in a list of 2-element tuples are decoded to
    unicode using UTF-8.
    """
    decoded = []
    for k, v in params:
        decoded.append((k.decode(u'utf-8') if isinstance(k, bytes) else k, v.decode(u'utf-8') if isinstance(v, bytes) else v))

    return decoded


urlencoded = set(always_safe) | set(u"=&;:%+~,*@!()/?'$")

def urldecode(query):
    u"""Decode a query string in x-www-form-urlencoded format into a sequence
    of two-element tuples.
    
    Unlike urlparse.parse_qsl(..., strict_parsing=True) urldecode will enforce
    correct formatting of the query string by validation. If validation fails
    a ValueError will be raised. urllib.parse_qsl will only raise errors if
    any of name-value pairs omits the equals sign.
    """
    if query and not set(query) <= urlencoded:
        error = u"Error trying to decode a non urlencoded string. Found invalid characters: %s in the string: '%s'. Please ensure the request/response body is x-www-form-urlencoded."
        raise ValueError(error % (set(query) - urlencoded, query))
    if INVALID_HEX_PATTERN.search(query):
        raise ValueError(u'Invalid hex encoding in query string.')
    query = query.encode(u'utf-8') if not PY3 and isinstance(query, unicode_type) else query
    params = urlparse.parse_qsl(query, keep_blank_values=True)
    return decode_params_utf8(params)


def extract_params(raw):
    u"""Extract parameters and return them as a list of 2-tuples.
    
    Will successfully extract parameters from urlencoded query strings,
    dicts, or lists of 2-tuples. Empty strings/dicts/lists will return an
    empty list of parameters. Any other input will result in a return
    value of None.
    """
    if isinstance(raw, bytes) or isinstance(raw, unicode_type):
        try:
            params = urldecode(raw)
        except ValueError:
            params = None

    elif hasattr(raw, u'__iter__'):
        try:
            dict(raw)
        except ValueError:
            params = None
        except TypeError:
            params = None
        else:
            params = list(raw.items() if isinstance(raw, dict) else raw)
            params = decode_params_utf8(params)

    else:
        params = None
    return params


def generate_nonce():
    u"""Generate pseudorandom nonce that is unlikely to repeat.
    
    Per `section 3.3`_ of the OAuth 1 RFC 5849 spec.
    Per `section 3.2.1`_ of the MAC Access Authentication spec.
    
    A random 64-bit number is appended to the epoch timestamp for both
    randomness and to decrease the likelihood of collisions.
    
    .. _`section 3.2.1`: https://tools.ietf.org/html/draft-ietf-oauth-v2-http-mac-01#section-3.2.1
    .. _`section 3.3`: https://tools.ietf.org/html/rfc5849#section-3.3
    """
    return unicode_type(unicode_type(randbits(64)) + generate_timestamp())


def generate_timestamp():
    u"""Get seconds since epoch (UTC).
    
    Per `section 3.3`_ of the OAuth 1 RFC 5849 spec.
    Per `section 3.2.1`_ of the MAC Access Authentication spec.
    
    .. _`section 3.2.1`: https://tools.ietf.org/html/draft-ietf-oauth-v2-http-mac-01#section-3.2.1
    .. _`section 3.3`: https://tools.ietf.org/html/rfc5849#section-3.3
    """
    return unicode_type(int(time.time()))


def generate_token(length = 30, chars = UNICODE_ASCII_CHARACTER_SET):
    u"""Generates a non-guessable OAuth token
    
    OAuth (1 and 2) does not specify the format of tokens except that they
    should be strings of random characters. Tokens should not be guessable
    and entropy when generating the random characters is important. Which is
    why SystemRandom is used instead of the default random.choice method.
    """
    rand = SystemRandom()
    return u''.join((rand.choice(chars) for x in range(length)))


def generate_signed_token(private_pem, request):
    import jwt
    now = datetime.datetime.utcnow()
    claims = {u'scope': request.scope,
     u'exp': now + datetime.timedelta(seconds=request.expires_in)}
    claims.update(request.claims)
    token = jwt.encode(claims, private_pem, u'RS256')
    token = to_unicode(token, u'UTF-8')
    return token


def verify_signed_token(public_pem, token):
    import jwt
    return jwt.decode(token, public_pem, algorithms=[u'RS256'])


def generate_client_id(length = 30, chars = CLIENT_ID_CHARACTER_SET):
    u"""Generates an OAuth client_id
    
    OAuth 2 specify the format of client_id in
    https://tools.ietf.org/html/rfc6749#appendix-A.
    """
    return generate_token(length, chars)


def add_params_to_qs(query, params):
    u"""Extend a query with a list of two-tuples."""
    if isinstance(params, dict):
        params = list(params.items())
    queryparams = urlparse.parse_qsl(query, keep_blank_values=True)
    queryparams.extend(params)
    return urlencode(queryparams)


def add_params_to_uri(uri, params, fragment = False):
    u"""Add a list of two-tuples to the uri query components."""
    sch, net, path, par, query, fra = urlparse.urlparse(uri)
    if fragment:
        fra = add_params_to_qs(fra, params)
    else:
        query = add_params_to_qs(query, params)
    return urlparse.urlunparse((sch,
     net,
     path,
     par,
     query,
     fra))


def safe_string_equals(a, b):
    u""" Near-constant time string comparison.
    
    Used in order to avoid timing attacks on sensitive information such
    as secret keys during request verification (`rootLabs`_).
    
    .. _`rootLabs`: http://rdist.root.org/2010/01/07/timing-independent-array-comparison/
    
    """
    if len(a) != len(b):
        return False
    result = 0
    for x, y in zip(a, b):
        result |= ord(x) ^ ord(y)

    return result == 0


def to_unicode(data, encoding = u'UTF-8'):
    u"""Convert a number of different types of objects to unicode."""
    if isinstance(data, unicode_type):
        return data
    if isinstance(data, bytes):
        return unicode_type(data, encoding=encoding)
    if hasattr(data, u'__iter__'):
        try:
            dict(data)
        except TypeError:
            pass
        except ValueError:
            return (to_unicode(i, encoding) for i in data)
        else:
            if hasattr(data, u'items'):
                data = list(data.items())
            return dict(((to_unicode(k, encoding), to_unicode(v, encoding)) for k, v in data))

    return data


class CaseInsensitiveDict(dict):
    u"""Basic case insensitive dict with strings only keys."""
    proxy = {}

    def __init__(self, data):
        self.proxy = dict(((k.lower(), k) for k in data))
        for k in data:
            self[k] = data[k]

    def __contains__(self, k):
        return k.lower() in self.proxy

    def __delitem__(self, k):
        key = self.proxy[k.lower()]
        super(CaseInsensitiveDict, self).__delitem__(key)
        del self.proxy[k.lower()]

    def __getitem__(self, k):
        key = self.proxy[k.lower()]
        return super(CaseInsensitiveDict, self).__getitem__(key)

    def get(self, k, default = None):
        if k in self:
            return self[k]
        return default

    def __setitem__(self, k, v):
        super(CaseInsensitiveDict, self).__setitem__(k, v)
        self.proxy[k.lower()] = k

    def update(self, *args, **kwargs):
        super(CaseInsensitiveDict, self).update(*args, **kwargs)
        for k in dict(*args, **kwargs):
            self.proxy[k.lower()] = k


class Request(object):
    u"""A malleable representation of a signable HTTP request.
    
    Body argument may contain any data, but parameters will only be decoded if
    they are one of:
    
    * urlencoded query string
    * dict
    * list of 2-tuples
    
    Anything else will be treated as raw body data to be passed through
    unmolested.
    """

    def __init__(self, uri, http_method = u'GET', body = None, headers = None, encoding = u'utf-8'):
        encode = lambda x: (to_unicode(x, encoding) if encoding else x)
        self.uri = encode(uri)
        self.http_method = encode(http_method)
        self.headers = CaseInsensitiveDict(encode(headers or {}))
        self.body = encode(body)
        self.decoded_body = extract_params(self.body)
        self.oauth_params = []
        self.validator_log = {}
        self._params = {u'access_token': None,
         u'client': None,
         u'client_id': None,
         u'client_secret': None,
         u'code': None,
         u'code_challenge': None,
         u'code_challenge_method': None,
         u'code_verifier': None,
         u'extra_credentials': None,
         u'grant_type': None,
         u'redirect_uri': None,
         u'refresh_token': None,
         u'request_token': None,
         u'response_type': None,
         u'scope': None,
         u'scopes': None,
         u'state': None,
         u'token': None,
         u'user': None,
         u'token_type_hint': None,
         u'response_mode': None,
         u'nonce': None,
         u'display': None,
         u'prompt': None,
         u'claims': None,
         u'max_age': None,
         u'ui_locales': None,
         u'id_token_hint': None,
         u'login_hint': None,
         u'acr_values': None}
        self._params.update(dict(urldecode(self.uri_query)))
        self._params.update(dict(self.decoded_body or []))
        return

    def __getattr__(self, name):
        if name in self._params:
            return self._params[name]
        raise AttributeError(name)

    def __repr__(self):
        body = self.body
        headers = self.headers.copy()
        if body:
            body = SANITIZE_PATTERN.sub(u'\x01<SANITIZED>', str(body))
        if u'Authorization' in headers:
            headers[u'Authorization'] = u'<SANITIZED>'
        return u'<oauthlib.Request url="%s", http_method="%s", headers="%s", body="%s">' % (self.uri,
         self.http_method,
         headers,
         body)

    @property
    def uri_query(self):
        return urlparse.urlparse(self.uri).query

    @property
    def uri_query_params(self):
        if not self.uri_query:
            return []
        return urlparse.parse_qsl(self.uri_query, keep_blank_values=True, strict_parsing=True)

    @property
    def duplicate_params(self):
        seen_keys = collections.defaultdict(int)
        all_keys = (p[0] for p in (self.decoded_body or []) + self.uri_query_params)
        for k in all_keys:
            seen_keys[k] += 1

        return [ k for k, c in list(seen_keys.items()) if c > 1 ]