u"""
oauthlib.oauth1.rfc5849.signature
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module represents a direct implementation of `section 3.4`_ of the spec.

Terminology:
 * Client: software interfacing with an OAuth API
 * Server: the API provider
 * Resource Owner: the user who is granting authorization to the client

Steps for signing a request:

1. Collect parameters from the uri query, auth header, & body
2. Normalize those parameters
3. Normalize the uri
4. Pass the normalized uri, normalized parameters, and http method to
   construct the base string
5. Pass the base string and any keys needed to a signing function

.. _`section 3.4`: https://tools.ietf.org/html/rfc5849#section-3.4
"""
from __future__ import absolute_import, unicode_literals
import binascii
import hashlib
import hmac
from . import common
from . import utils
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

def construct_base_string(http_method, base_string_uri, normalized_encoded_request_parameters):
    u"""**String Construction**
    Per `section 3.4.1.1`_ of the spec.
    
    For example, the HTTP request::
    
        POST /request?b5=%3D%253D&a3=a&c%40=&a2=r%20b HTTP/1.1
        Host: example.com
        Content-Type: application/x-www-form-urlencoded
        Authorization: OAuth realm="Example",
            oauth_consumer_key="9djdj82h48djs9d2",
            oauth_token="kkk9d7dh3k39sjv7",
            oauth_signature_method="HMAC-SHA1",
            oauth_timestamp="137131201",
            oauth_nonce="7d8f3e4a",
            oauth_signature="bYT5CMsGcbgUdFHObYMEfcx6bsw%3D"
    
        c2&a3=2+q
    
    is represented by the following signature base string (line breaks
    are for display purposes only)::
    
        POST&http%3A%2F%2Fexample.com%2Frequest&a2%3Dr%2520b%26a3%3D2%2520q
        %26a3%3Da%26b5%3D%253D%25253D%26c%2540%3D%26c2%3D%26oauth_consumer_
        key%3D9djdj82h48djs9d2%26oauth_nonce%3D7d8f3e4a%26oauth_signature_m
        ethod%3DHMAC-SHA1%26oauth_timestamp%3D137131201%26oauth_token%3Dkkk
        9d7dh3k39sjv7
    
    .. _`section 3.4.1.1`: https://tools.ietf.org/html/rfc5849#section-3.4.1.1
    """
    base_string = utils.escape(http_method.upper())
    base_string += u'&'
    base_string += utils.escape(base_string_uri)
    base_string += u'&'
    base_string += utils.escape(normalized_encoded_request_parameters)
    return base_string


def normalize_base_string_uri(uri, host = None):
    u"""**Base String URI**
    Per `section 3.4.1.2`_ of the spec.
    
    For example, the HTTP request::
    
        GET /r%20v/X?id=123 HTTP/1.1
        Host: EXAMPLE.COM:80
    
    is represented by the base string URI: "http://example.com/r%20v/X".
    
    In another example, the HTTPS request::
    
        GET /?q=1 HTTP/1.1
        Host: www.example.net:8080
    
    is represented by the base string URI: "https://www.example.net:8080/".
    
    .. _`section 3.4.1.2`: https://tools.ietf.org/html/rfc5849#section-3.4.1.2
    
    The host argument overrides the netloc part of the uri argument.
    """
    if not isinstance(uri, common.unicode_type):
        raise ValueError(u'uri must be a unicode object.')
    scheme, netloc, path, params, query, fragment = urlparse.urlparse(uri)
    if not scheme or not netloc:
        raise ValueError(u'uri must include a scheme and netloc')
    if not path:
        path = u'/'
    scheme = scheme.lower()
    netloc = netloc.lower()
    if host is not None:
        netloc = host.lower()
    default_ports = ((u'http', u'80'), (u'https', u'443'))
    if u':' in netloc:
        host, port = netloc.split(u':', 1)
        if (scheme, port) in default_ports:
            netloc = host
    return urlparse.urlunparse((scheme,
     netloc,
     path,
     params,
     u'',
     u''))


def collect_parameters(uri_query = u'', body = [], headers = None, exclude_oauth_signature = True, with_realm = False):
    u"""**Parameter Sources**
    
    Parameters starting with `oauth_` will be unescaped.
    
    Body parameters must be supplied as a dict, a list of 2-tuples, or a
    formencoded query string.
    
    Headers must be supplied as a dict.
    
    Per `section 3.4.1.3.1`_ of the spec.
    
    For example, the HTTP request::
    
        POST /request?b5=%3D%253D&a3=a&c%40=&a2=r%20b HTTP/1.1
        Host: example.com
        Content-Type: application/x-www-form-urlencoded
        Authorization: OAuth realm="Example",
            oauth_consumer_key="9djdj82h48djs9d2",
            oauth_token="kkk9d7dh3k39sjv7",
            oauth_signature_method="HMAC-SHA1",
            oauth_timestamp="137131201",
            oauth_nonce="7d8f3e4a",
            oauth_signature="djosJKDKJSD8743243%2Fjdk33klY%3D"
    
        c2&a3=2+q
    
    contains the following (fully decoded) parameters used in the
    signature base sting::
    
        +------------------------+------------------+
        |          Name          |       Value      |
        +------------------------+------------------+
        |           b5           |       =%3D       |
        |           a3           |         a        |
        |           c@           |                  |
        |           a2           |        r b       |
        |   oauth_consumer_key   | 9djdj82h48djs9d2 |
        |       oauth_token      | kkk9d7dh3k39sjv7 |
        | oauth_signature_method |     HMAC-SHA1    |
        |     oauth_timestamp    |     137131201    |
        |       oauth_nonce      |     7d8f3e4a     |
        |           c2           |                  |
        |           a3           |        2 q       |
        +------------------------+------------------+
    
    Note that the value of "b5" is "=%3D" and not "==".  Both "c@" and
    "c2" have empty values.  While the encoding rules specified in this
    specification for the purpose of constructing the signature base
    string exclude the use of a "+" character (ASCII code 43) to
    represent an encoded space character (ASCII code 32), this practice
    is widely used in "application/x-www-form-urlencoded" encoded values,
    and MUST be properly decoded, as demonstrated by one of the "a3"
    parameter instances (the "a3" parameter is used twice in this
    request).
    
    .. _`section 3.4.1.3.1`: https://tools.ietf.org/html/rfc5849#section-3.4.1.3.1
    """
    headers = headers or {}
    params = []
    if uri_query:
        params.extend(common.urldecode(uri_query))
    if headers:
        headers_lower = dict(((k.lower(), v) for k, v in list(headers.items())))
        authorization_header = headers_lower.get(u'authorization')
        if authorization_header is not None:
            params.extend([ i for i in utils.parse_authorization_header(authorization_header) if with_realm or i[0] != u'realm' ])
    bodyparams = common.extract_params(body) or []
    params.extend(bodyparams)
    unescaped_params = []
    for k, v in params:
        if k.startswith(u'oauth_'):
            v = utils.unescape(v)
        unescaped_params.append((k, v))

    if exclude_oauth_signature:
        unescaped_params = list([i for i in unescaped_params if i[0] != 'oauth_signature'])
    return unescaped_params


def normalize_parameters(params):
    u"""**Parameters Normalization**
    Per `section 3.4.1.3.2`_ of the spec.
    
    For example, the list of parameters from the previous section would
    be normalized as follows:
    
    Encoded::
    
    +------------------------+------------------+
    |          Name          |       Value      |
    +------------------------+------------------+
    |           b5           |     %3D%253D     |
    |           a3           |         a        |
    |          c%40          |                  |
    |           a2           |       r%20b      |
    |   oauth_consumer_key   | 9djdj82h48djs9d2 |
    |       oauth_token      | kkk9d7dh3k39sjv7 |
    | oauth_signature_method |     HMAC-SHA1    |
    |     oauth_timestamp    |     137131201    |
    |       oauth_nonce      |     7d8f3e4a     |
    |           c2           |                  |
    |           a3           |       2%20q      |
    +------------------------+------------------+
    
    Sorted::
    
    +------------------------+------------------+
    |          Name          |       Value      |
    +------------------------+------------------+
    |           a2           |       r%20b      |
    |           a3           |       2%20q      |
    |           a3           |         a        |
    |           b5           |     %3D%253D     |
    |          c%40          |                  |
    |           c2           |                  |
    |   oauth_consumer_key   | 9djdj82h48djs9d2 |
    |       oauth_nonce      |     7d8f3e4a     |
    | oauth_signature_method |     HMAC-SHA1    |
    |     oauth_timestamp    |     137131201    |
    |       oauth_token      | kkk9d7dh3k39sjv7 |
    +------------------------+------------------+
    
    Concatenated Pairs::
    
    +-------------------------------------+
    |              Name=Value             |
    +-------------------------------------+
    |               a2=r%20b              |
    |               a3=2%20q              |
    |                 a3=a                |
    |             b5=%3D%253D             |
    |                c%40=                |
    |                 c2=                 |
    | oauth_consumer_key=9djdj82h48djs9d2 |
    |         oauth_nonce=7d8f3e4a        |
    |   oauth_signature_method=HMAC-SHA1  |
    |      oauth_timestamp=137131201      |
    |     oauth_token=kkk9d7dh3k39sjv7    |
    +-------------------------------------+
    
    and concatenated together into a single string (line breaks are for
    display purposes only)::
    
        a2=r%20b&a3=2%20q&a3=a&b5=%3D%253D&c%40=&c2=&oauth_consumer_key=9dj
        dj82h48djs9d2&oauth_nonce=7d8f3e4a&oauth_signature_method=HMAC-SHA1
        &oauth_timestamp=137131201&oauth_token=kkk9d7dh3k39sjv7
    
    .. _`section 3.4.1.3.2`: https://tools.ietf.org/html/rfc5849#section-3.4.1.3.2
    """
    key_values = [ (utils.escape(k), utils.escape(v)) for k, v in params ]
    key_values.sort()
    parameter_parts = [ u'{0}={1}'.format(k, v) for k, v in key_values ]
    return u'&'.join(parameter_parts)


def sign_hmac_sha1_with_client(base_string, client):
    return sign_hmac_sha1(base_string, client.client_secret, client.resource_owner_secret)


def sign_hmac_sha1(base_string, client_secret, resource_owner_secret):
    u"""**HMAC-SHA1**
    
    The "HMAC-SHA1" signature method uses the HMAC-SHA1 signature
    algorithm as defined in `RFC2104`_::
    
        digest = HMAC-SHA1 (key, text)
    
    Per `section 3.4.2`_ of the spec.
    
    .. _`RFC2104`: https://tools.ietf.org/html/rfc2104
    .. _`section 3.4.2`: https://tools.ietf.org/html/rfc5849#section-3.4.2
    """
    text = base_string
    key = utils.escape(client_secret or u'')
    key += u'&'
    key += utils.escape(resource_owner_secret or u'')
    key_utf8 = key.encode(u'utf-8')
    text_utf8 = text.encode(u'utf-8')
    signature = hmac.new(key_utf8, text_utf8, hashlib.sha1)
    return binascii.b2a_base64(signature.digest())[:-1].decode(u'utf-8')


def sign_hmac_sha256_with_client(base_string, client):
    return sign_hmac_sha256(base_string, client.client_secret, client.resource_owner_secret)


def sign_hmac_sha256(base_string, client_secret, resource_owner_secret):
    u"""**HMAC-SHA256**
    
    The "HMAC-SHA256" signature method uses the HMAC-SHA256 signature
    algorithm as defined in `RFC4634`_::
    
        digest = HMAC-SHA256 (key, text)
    
    Per `section 3.4.2`_ of the spec.
    
    .. _`RFC4634`: https://tools.ietf.org/html/rfc4634
    .. _`section 3.4.2`: https://tools.ietf.org/html/rfc5849#section-3.4.2
    """
    text = base_string
    key = utils.escape(client_secret or u'')
    key += u'&'
    key += utils.escape(resource_owner_secret or u'')
    key_utf8 = key.encode(u'utf-8')
    text_utf8 = text.encode(u'utf-8')
    signature = hmac.new(key_utf8, text_utf8, hashlib.sha256)
    return binascii.b2a_base64(signature.digest())[:-1].decode(u'utf-8')


_jwtrs1 = None

def _jwt_rs1_signing_algorithm():
    global _jwtrs1
    if _jwtrs1 is None:
        import jwt.algorithms as jwtalgo
        _jwtrs1 = jwtalgo.RSAAlgorithm(jwtalgo.hashes.SHA1)
    return _jwtrs1


def sign_rsa_sha1(base_string, rsa_private_key):
    u"""**RSA-SHA1**
    
    Per `section 3.4.3`_ of the spec.
    
    The "RSA-SHA1" signature method uses the RSASSA-PKCS1-v1_5 signature
    algorithm as defined in `RFC3447, Section 8.2`_ (also known as
    PKCS#1), using SHA-1 as the hash function for EMSA-PKCS1-v1_5.  To
    use this method, the client MUST have established client credentials
    with the server that included its RSA public key (in a manner that is
    beyond the scope of this specification).
    
    .. _`section 3.4.3`: https://tools.ietf.org/html/rfc5849#section-3.4.3
    .. _`RFC3447, Section 8.2`: https://tools.ietf.org/html/rfc3447#section-8.2
    
    """
    if isinstance(base_string, common.unicode_type):
        base_string = base_string.encode(u'utf-8')
    alg = _jwt_rs1_signing_algorithm()
    key = _prepare_key_plus(alg, rsa_private_key)
    s = alg.sign(base_string, key)
    return binascii.b2a_base64(s)[:-1].decode(u'utf-8')


def sign_rsa_sha1_with_client(base_string, client):
    if not client.rsa_key:
        raise ValueError(u'rsa_key is required when using RSA signature method.')
    return sign_rsa_sha1(base_string, client.rsa_key)


def sign_plaintext(client_secret, resource_owner_secret):
    u"""Sign a request using plaintext.
    
    Per `section 3.4.4`_ of the spec.
    
    The "PLAINTEXT" method does not employ a signature algorithm.  It
    MUST be used with a transport-layer mechanism such as TLS or SSL (or
    sent over a secure channel with equivalent protections).  It does not
    utilize the signature base string or the "oauth_timestamp" and
    "oauth_nonce" parameters.
    
    .. _`section 3.4.4`: https://tools.ietf.org/html/rfc5849#section-3.4.4
    
    """
    signature = utils.escape(client_secret or u'')
    signature += u'&'
    signature += utils.escape(resource_owner_secret or u'')
    return signature


def sign_plaintext_with_client(base_string, client):
    return sign_plaintext(client.client_secret, client.resource_owner_secret)


def verify_hmac_sha1(request, client_secret = None, resource_owner_secret = None):
    u"""Verify a HMAC-SHA1 signature.
    
    Per `section 3.4`_ of the spec.
    
    .. _`section 3.4`: https://tools.ietf.org/html/rfc5849#section-3.4
    
    To satisfy `RFC2616 section 5.2`_ item 1, the request argument's uri
    attribute MUST be an absolute URI whose netloc part identifies the
    origin server or gateway on which the resource resides. Any Host
    item of the request argument's headers dict attribute will be
    ignored.
    
    .. _`RFC2616 section 5.2`: https://tools.ietf.org/html/rfc2616#section-5.2
    
    """
    norm_params = normalize_parameters(request.params)
    uri = normalize_base_string_uri(request.uri)
    base_string = construct_base_string(request.http_method, uri, norm_params)
    signature = sign_hmac_sha1(base_string, client_secret, resource_owner_secret)
    match = common.safe_string_equals(signature, request.signature)
    return match


def _prepare_key_plus(alg, keystr):
    if isinstance(keystr, bytes):
        keystr = keystr.decode(u'utf-8')
    return alg.prepare_key(keystr)


def verify_rsa_sha1(request, rsa_public_key):
    u"""Verify a RSASSA-PKCS #1 v1.5 base64 encoded signature.
    
    Per `section 3.4.3`_ of the spec.
    
    Note this method requires the jwt and cryptography libraries.
    
    .. _`section 3.4.3`: https://tools.ietf.org/html/rfc5849#section-3.4.3
    
    To satisfy `RFC2616 section 5.2`_ item 1, the request argument's uri
    attribute MUST be an absolute URI whose netloc part identifies the
    origin server or gateway on which the resource resides. Any Host
    item of the request argument's headers dict attribute will be
    ignored.
    
    .. _`RFC2616 section 5.2`: https://tools.ietf.org/html/rfc2616#section-5.2
    """
    norm_params = normalize_parameters(request.params)
    uri = normalize_base_string_uri(request.uri)
    message = construct_base_string(request.http_method, uri, norm_params).encode(u'utf-8')
    sig = binascii.a2b_base64(request.signature.encode(u'utf-8'))
    alg = _jwt_rs1_signing_algorithm()
    key = _prepare_key_plus(alg, rsa_public_key)
    verify_ok = alg.verify(message, key, sig)
    return verify_ok


def verify_plaintext(request, client_secret = None, resource_owner_secret = None):
    u"""Verify a PLAINTEXT signature.
    
    Per `section 3.4`_ of the spec.
    
    .. _`section 3.4`: https://tools.ietf.org/html/rfc5849#section-3.4
    """
    signature = sign_plaintext(client_secret, resource_owner_secret)
    match = common.safe_string_equals(signature, request.signature)
    return match